/* Approximate Cover Listing Algorithm */
/* 2008/1/20 Takeaki Uno,   e-mail:uno@nii.jp, 
    homepage:   http://research.nii.ac.jp/~uno/index.html  */
/* This program is available for only academic use, basically.
   Anyone can modify this program, but he/she has to write down 
    the change of the modification on the top of the source code.
   Neither contact nor appointment to Takeaki Uno is needed.
   If one wants to re-distribute this code, please 
    refer the newest code, and show the link to homepage of 
    Takeaki Uno, to notify the news about ACLA for the users. */


#ifndef _acla_c_
#define _acla_c_

#define WEIGHT_DOUBLE

#include"trsact.c"
#include"problem.c"

void ACLA_error (){
  ERROR_MES = "command explanation";
  print_err ("ACLA: [FfQIq] [options] input-filename max_size support [output-filename]\n\
%%:show progress, _:no message, +:write solutions in append mode\n\
F:approximate cover mining, C:irredandunt approximate cover mining\n\
f,Q:output coverage following/preceding to each output itemset\n\
A:output coverages for positive/negative transactions\n\
I:output ID's of transactions including each pattern\n\
t:transpose the database so that i-th transaction will be item i\n\
[options]\n\
-K [num]:output [num] most frequent itemsets\n\
-l [num]:output itemsets with size at least [num]\n\
-u [num]:output itemsets with size at most [num]\n\
-U [num]:upper bound for support(maximum support)\n\
-w [filename]:read weights of transactions from the file\n\
-N,-n [num]:give upper/lower bound for negative weight sum (frequency)\n\
-O,-o [num]:give upper/lower bound for positive weight sum (frequency)\n\
-# [num]:stop after outputting [num] solutions\n\
-, [char]:give the separator of the numbers in the output\n\
-Q [filename]:replace the output numbers according to the permutation table given by [filename]\n\
# the 1st letter of input-filename cannot be '-'.\n\
# if the output file name is -, the solutions will be output to standard output.\n");
  EXIT;
}

/***********************************************************************/
/*  read parameters given by command line  */
/***********************************************************************/
void ACLA_read_param (int argc, char *argv[], PROBLEM *PP){
  int c=1;
  ITEMSET *II = &PP->II;
  if ( argc < c+4 ){ ACLA_error (); return; }

  if ( !strchr (argv[c], '_') ){ II->flag |= SHOW_MESSAGE; PP->TT.flag |= SHOW_MESSAGE; }
  if ( strchr( argv[c], '%' ) ) II->flag |= SHOW_PROGRESS;
  if ( strchr (argv[c], '+') ) II->flag |= ITEMSET_APPEND;
  if ( strchr (argv[c], 'f') ) II->flag |= ITEMSET_FREQ;
  if ( strchr (argv[c], 'Q') ) II->flag |= ITEMSET_PRE_FREQ;
  if ( strchr (argv[c], 'A') ) II->flag |= ITEMSET_OUTPUT_POSINEGA;
  if ( strchr (argv[c], 'F') ){ PP->problem |= PROBLEM_FREQSET; II->flag |= ITEMSET_ALL; }
  else if ( strchr( argv[c], 'C' ) ){ PP->problem |= PROBLEM_CLOSED; }
  else error ("F or C command has to be specified", EXIT);
  if ( strchr( argv[c], 'I' ) ) II->flag |= ITEMSET_TRSACT_ID;
  if ( strchr (argv[c], 't') ) PP->TT.flag |= LOAD_TPOSE;
  c++;
  
  while ( argv[c][0] == '-' ){
    switch (argv[c][1]){
      case 'K': II->topk.end = atoi (argv[c+1]);
      break; case 'l': II->lb = atoi (argv[c+1]);
      break; case 'u': II->ub = atoi(argv[c+1]);
      break; case 'U': II->frq_ub = (WEIGHT)atof(argv[c+1]);
      break; case 'w': PP->TT.wfname = argv[c+1];
      break; case 'n': II->nega_lb = atof(argv[c+1]);
      break; case 'N': II->nega_ub = atof(argv[c+1]);
      break; case 'o': II->posi_lb = atof(argv[c+1]);
      break; case 'O': II->posi_ub = atof(argv[c+1]);
      break; case '#': II->max_solutions = atoi(argv[c+1]);
      break; case ',': II->separator = argv[c+1][0];
      break; case 'Q': PP->outperm_fname = argv[c+1];
      break; default: goto NEXT;
    }
    c += 2;
    if ( argc < c+3 ){ ACLA_error (); return; }
  }

  NEXT:;
  PP->TT.fname = argv[c];
  II->ub = atoi(argv[c+1]);
  if ( II->topk.end==0 ) II->frq_lb = (WEIGHT)atof(argv[c+2]);
  if ( argc>c+3 ) PP->output_fname = argv[c+3];
}

/***************************************************************/
/* iteration of ACLA */
/* INPUT: item:tail of the current solution, t_new,buf:head of the list of 
 ID and buffer memory of new transactions */
/* (flag&4) => bitmap ppc mode */
/*************************************************************************/
void ACLA (PROBLEM *PP, QUEUE_INT item, QUEUE *occ, WEIGHT sum, WEIGHT psum){
  ITEMSET *II = &PP->II;
  TRSACT *TT = &PP->TT;
  QUEUE_INT e, *x;
  VEC_ID new_t = TT->new_t;
  QUEUE_ID js = PP->itemjump.s;
  QUEUE occ2;
  int bnum = TT->buf.num, bblock_num = TT->buf.block_num;
  WEIGHT m1=-WEIGHTHUGE, m1_=WEIGHTHUGE, n;
  WEIGHT p1=0, p2=0, p3=0, p1_=0;
  WEIGHT n1=0, n2=0, n3=0, n1_=0;

  II->iters++;
  if ( II->itemset.t == II->ub || ( (PP->problem & PROBLEM_CLOSED) && sum >= II->frq_lb ) ){
    II->frq = sum; II->pfrq = psum;
    ITEMSET_output_itemset (II, occ, 0);
    return;
  }

  PP->itemjump.s = PP->itemjump.t; // initilization for the re-use of queue
  TRSACT_delivery (TT, &PP->itemjump, PP->occ_w, PP->occ_pw, occ, item);
  if ( PP->itemjump.s == PP->itemjump.t ) goto END;

    // output routine for the case that |itemset| == goal-1
  if ( II->itemset.t == II->ub-1 && (II->flag & ITEMSET_TRSACT_ID)==0){
    MQUE_FLOOP (PP->itemjump, x){
      n = (sum-psum) + PP->occ_w[*x] - PP->occ_pw[*x];
      if ( sum+PP->occ_w[*x] >= II->frq_lb && sum+PP->occ_w[*x] <= II->frq_ub 
        && psum + PP->occ_pw[*x] >= II->posi_lb && psum + PP->occ_pw[*x] <= II->posi_ub
        &&  n >= II->nega_lb &&  n <= II->nega_ub ){
          II->frq = sum + PP->occ_w[*x];
          II->pfrq = psum + PP->occ_pw[*x];
          QUE_INS (II->itemset, *x);
          ITEMSET_output_itemset (II, occ, 0);
          II->itemset.t--;
      }
    }
    goto END;
  }

   // compute top-3 most heavy/light items
  MQUE_FLOOP (PP->itemjump, x){
    n = PP->occ_w[*x] - PP->occ_pw[*x];
    if ( PP->occ_w[*x] > m1 ){ m1=PP->occ_w[*x]; }
    if ( PP->occ_pw[*x] > p1 ){ p3=p2; p2=p1; p1=PP->occ_pw[*x]; }
    else if ( PP->occ_pw[*x] > p2 ){ p3=p2; p2=PP->occ_pw[*x]; }
    else if ( PP->occ_pw[*x] > p3 ){ p3=PP->occ_pw[*x]; }
    if ( n < n1 ){ n3=n2; n2=n1; n1=n; }
    else if ( n < n2 ){ n3=n2; n2=n; }
    else if ( n < n3 ){ n3=n; }

    if ( PP->occ_w[*x] < m1_ ){ m1_=PP->occ_w[*x]; }
    if ( PP->occ_pw[*x] < p1_ ){ p1_=PP->occ_pw[*x]; }
    if ( n > n1_ ){ n1_=n; }
  }

    // pruning by most heavy/light item for upper bounds
  if ( m1_+n1_*(II->ub-II->itemset.t-1) > II->frq_ub-sum ) goto END;
  if ( p1_ > II->posi_ub-psum ) goto END;
  if ( n1_ < II->nega_lb-(sum-psum) ) goto END;

    // pruning by most heavy/light 3 items for lower bounds
  if ( II->itemset.t == II->ub-2 && m1+p2 < II->frq_lb-sum ) goto END;
  else if ( m1+p2+p3*(II->ub-II->itemset.t-2) < II->frq_lb-sum ) goto END;
  if ( II->itemset.t == II->ub-2 && p1+p2 < II->posi_lb-psum ) goto END;
  else if ( p1+p2+p3*(II->ub-II->itemset.t-2) < II->posi_lb-psum ) goto END;
  if ( II->itemset.t == II->ub-2 && n1+n2 > II->nega_ub-(sum-psum) ) goto END;
  else if ( n1+n2+n3*(II->ub-II->itemset.t-2) > II->nega_ub-(sum-psum) ) goto END;

  qsort_QUEUE_INT (PP->itemjump.v, PP->itemjump.t, -1);

    // database reduction
  if ( II->itemset.t >0 && QUEUE_LENGTH_(PP->itemjump)>2 && (II->flag & ITEMSET_TRSACT_ID)==0 ){
    TRSACT_find_same (TT, occ, item);
    TRSACT_merge_trsact (TT, &TT->OQ[TT->T.clms], item);
    TRSACT_reduce_occ (TT, occ);
  }
    // occurrence deliver
  TRSACT_deliv (TT, occ, item);
  QUEUE_alloc (&occ2, occ->t+1);
    // loop for recursive calls
  while ( QUEUE_LENGTH_(PP->itemjump) > 0 ){
    e = QUEUE_ext_tail_ (&PP->itemjump);
    QUE_INS (II->itemset, e);

     // take complement of the occurrence
    occ2.t = occ2.s = 0;
    MQUE_FLOOP (*occ, x) TT->mark[*x] = 1;
    MQUE_FLOOP (TT->OQ[e], x) TT->mark[*x] = 0;
    MQUE_FLOOP (*occ, x) if ( TT->mark[*x] ) QUE_INS (occ2, *x);

      ACLA (PP, e, &occ2, sum + PP->occ_w[e], psum + PP->occ_pw[e]);

    II->itemset.t--;
    TT->OQ[e].t = TT->OQ[e].end = 0;
    PP->occ_w[e] = PP->occ_pw[e] = -WEIGHTHUGE;
  }

  TT->new_t = new_t;
  TT->buf.num = bnum; TT->buf.block_num = bblock_num;
  QUEUE_end (&occ2);

  END:;
  MQUE_FLOOP (PP->itemjump, x) TT->OQ[*x].t = TT->OQ[*x].end = 0;
  PP->itemjump.t = PP->itemjump.s;
  PP->itemjump.s = js;
}

/*************************************************************************/
/* main of ACLA */
/*************************************************************************/
int ACLA_main ( int argc, char *argv[] ){
  PROBLEM PP;
  TRSACT *TT = &PP.TT;
  
  PROBLEM_init ( &PP);
  ACLA_read_param ( argc, argv, &PP);
if ( ERROR_MES ) return (1);

  TT->flag |= LOAD_INCSORT +TRSACT_FRQSORT +LOAD_RM_DUP +TRSACT_SHRINK +TRSACT_1ST_SHRINK +TRSACT_MAKE_NEW;
  TT->th = -WEIGHTHUGE;
  
  PROBLEM_load (&PP);
  PROBLEM_alloc (&PP, TT->T.clms, TT->T.t, TT->T.eles, TT->perm, ((TT->flag&TRSACT_NEGATIVE)?PROBLEM_OCC_PW: PROBLEM_OCC_W) +PROBLEM_ITEMJUMP);
  PP.oo = QUEUE_dup_ (&TT->OQ[TT->T.clms]);  // preserve occ
  TT->perm = NULL;
  TT->OQ[TT->T.clms].t = 0;
  print_mes (&PP.TT, "separated at %d\n", PP.TT.sep);
  if ( !ERROR_MES ){
    ACLA (&PP, TT->T.clms, &PP.oo, 0, 0);
    ITEMSET_last_output (&PP.II);
  }
  PROBLEM_end (&PP);
  return (ERROR_MES?1:0);
}

/*******************************************************************************/
#ifndef _NO_MAIN_
#define _NO_MAIN_
int main (int argc, char *argv[]){
  return (ACLA_main (argc, argv));
}
#endif
/*******************************************************************************/

#endif

